;; Generated package description from systemd.el  -*- no-byte-compile: t -*-
(define-package "systemd" "0.0.20221212.230647" "Interface to Systemd" '((cl-lib "0.5")) :commit "807d8ce3df8cb4cf16de919f6edfcdbe4dbb3469" :url "https://elpa.gnu.org/packages/systemd.html" :authors '(("Mario Lang" . "mlang@delysid.org")) :maintainer '("Mario Lang" . "mlang@delysid.org") :keywords '("comm"))
